/*
 * romfs.c - Simple code to locate files within the ROM-based filesystem.
 *
 * Copyright (c) 2003-2004 Marcus R. Brown <mrbrown@0xd6.org>
 *
 * Licensed under the AFL v2.0. See the file LICENSE included with this
 * distribution for licensing terms.
 */

#include "types.h"
#include "defs.h"

#define BIOS_START (void *)0xbfc00000
#define BIOS_END   (void *)0xbfc10000

struct romfs_info {
	void	*base;		/* Base of ROM */
	void	*start;		/* Start of the root directory */
	void	*end;		/* End of the root directory */
};

struct romfs_dirent {
	u8	name[10];
	u16	ext_size;	/* Size of the extended info */
	u32	size;		/* Size in bytes */
};

static struct romfs_info root_info = { NULL, NULL, NULL };

/* Compare 10 bytes of a romfs filename */
static inline int compare10(void *src, const void *dest)
{
	u32 *s = (u32 *)src, *d = (u32 *)dest;

	if ((s[0] == d[0]) && (s[1] == d[1]) &&
			(_lh((u32)&s[2]) == _lh((u32)&d[2])))
		return 1;

	return 0;
}

/*
 * Find the romfs root directory, given a range of addresses to check.
 * It returns a pointer to information on the location of the romfs root dir,
 */
static struct romfs_info * root_search(void *start, void *end)
{
	struct romfs_info *root_dir = &root_info;
	struct romfs_dirent *dirp = (struct romfs_dirent *)start;
	u32 pos = 0;

	if (root_dir->base)
		return root_dir;

	do {
		/* 
		 * Look for the "RESET" file, which indicates the beginning of
		 * the root directory.
		 */
		if (compare10(&dirp->name[0], "RESET\0\0\0\0") &&
				(ALIGN(dirp->size, 16) == pos)) {
			root_dir->base  = start;
			root_dir->start = dirp;
			root_dir->end   = dirp + dirp->size;
			return root_dir;
		}
		dirp++;
		pos += 16;
	} while ((void *)dirp < end);

	return NULL;
}

/*
 * Returns the starting address, size, and address of the extended information
 * of the given filename.  Returns 1 on success, 0 if the file isn't found.
 *
 * The passed-in filename must be padded out to 10 bytes with the '\0' (nul)
 * character.
 */
int romfs_findfile(const char *filename, u32 *addr, u32 *size,
		u32 *ext_addr)
{
	struct romfs_info *root = root_search(BIOS_START, BIOS_END);
	struct romfs_dirent *dirp;
	u32 curofs = 0, extofs = 0;

	if (!root)
		return 0;

	dirp = (struct romfs_dirent *)root->start;
	do {
		if (compare10(&dirp->name, filename)) {

			if (addr)
				*addr = (u32)root->base + curofs;
			if (size)
				*size = dirp->size;
			if (dirp->ext_size && ext_addr)
				*ext_addr = (u32)root->end + extofs;

			return 1;
		}

		curofs += ALIGN(dirp->size, 16);
		extofs += dirp->ext_size;
		dirp++;
	} while (dirp->name[0]);

	return 0;
}
